/* vi:set ts=8 sts=4 sw=4:
 * VIM - Vi IMproved            by Bram Moolenaar
 *
 * Do ":help uganda"  in Vim to read copying and usage conditions.
 * Do ":help credits" in Vim to see a list of people who contributed.
 */

/*
 * Original porting to GTK+:
 * (C) 1998 by Marcin Dalecki <dalecki@cs.net.pl>
 *
 * Anybody else please just add yourself here...
 */

#include "vim.h"

#include <gdk/gdkkeysyms.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>

#include <gtk/gtk.h>

#include <X11/Xatom.h>

#include "gui_gtk_f.h"

#define VIM_NAME	"vim"
#define VIM_CLASS	"Vim"

/* slection distinguishers */
enum {
    SEL_TYPE_NONE,
    SELECTION_STRING,
    SELECTION_CLIPBOARD
};

/* This is the single only fixed width font in X11, which seems to be present
 * on all servers and available in all the variants we need.
 *
 * Don't try to tell me that X11 is a wonderfull technology.  If You need to
 * look for a nice GUI design look for the PalmOS instead!
 */

#define DFLT_FONT		"-adobe-courier-medium-r-normal-*-14-*-*-*-m-*-*-*"

GtkWidget *main_window = NULL;
static GtkWidget *main_vbox = NULL;
GtkWidget *form_window = NULL;
GtkWidget *drawing_area = NULL;
GtkWidget *menu_bar = NULL;
GtkWidget *tool_bar = NULL;
#ifdef GTK_HAVE_FEATURES_1_1_0
GtkAccelGroup *accel_group = NULL;

/* This is used to pass the result from the dialog to the upper layers. */
static char_u *selected_font_name = NULL;
#endif

static GdkColor foreground_color;
static GdkColor background_color;

#define DELETE_WINDOW_IDX 0	/* index in wm_atoms[] for WM_DELETE_WINDOW */
#define SAVE_YOURSELF_IDX 1	/* index in wm_atoms[] for WM_SAVE_YOURSELF */

#ifdef GTK_HAVE_FEATURES_1_1_0
static void font_selection_ok(GtkWidget * w, GtkFontSelectionDialog * fs);
static void font_selection_destroy(GtkWidget * w, gpointer * data);
#endif
static gint expose_event(GtkWidget * widget, GdkEventExpose * event);
static gint key_press_event(GtkWidget *, GdkEventKey * event, gpointer data);
static gint button_press_event(GtkWidget * widget, GdkEventButton * event);
static gint button_release_event(GtkWidget * widget, GdkEventButton * event);
static gint motion_notify_event(GtkWidget * widget, GdkEventMotion * event);
static void destroy_callback(void);

/*
 * Keycodes recognized by vim.
 * we will be using the techniques described here later for real work.
 */
static struct special_key {
    guint key_sym;
    char_u code0;
    char_u code1;
} special_keys[] =
{
    { GDK_Up,		'k', 'u' },
    { GDK_Down,		'k', 'd' },
    { GDK_Left,		'k', 'l' },
    { GDK_Right,	'k', 'r' },
    { GDK_F1,		'k', '1' },
    { GDK_F2,		'k', '2' },
    { GDK_F3,		'k', '3' },
    { GDK_F4,		'k', '4' },
    { GDK_F5,		'k', '5' },
    { GDK_F6,		'k', '6' },
    { GDK_F7,		'k', '7' },
    { GDK_F8,		'k', '8' },
    { GDK_F9,		'k', '9' },
    { GDK_F10,		'k', ';' },
    { GDK_F11,		'F', '1' },
    { GDK_F12,		'F', '2' },
    { GDK_F13,		'F', '3' },
    { GDK_F14,		'F', '4' },
    { GDK_F15,		'F', '5' },
    { GDK_F16,		'F', '6' },
    { GDK_F17,		'F', '7' },
    { GDK_F18,		'F', '8' },
    { GDK_F19,		'F', '9' },
    { GDK_F20,		'F', 'A' },
    { GDK_F21,		'F', 'B' },
    { GDK_F22,		'F', 'C' },
    { GDK_F23,		'F', 'D' },
    { GDK_F24,		'F', 'E' },
    { GDK_F25,		'F', 'F' },
    { GDK_F26,		'F', 'G' },
    { GDK_F27,		'F', 'H' },
    { GDK_F28,		'F', 'I' },
    { GDK_F29,		'F', 'J' },
    { GDK_F30,		'F', 'K' },
    { GDK_F31,		'F', 'L' },
    { GDK_F32,		'F', 'M' },
    { GDK_F33,		'F', 'N' },
    { GDK_F34,		'F', 'O' },
    { GDK_F35,		'F', 'P' },
    { GDK_Help,		'%', '1' },
    { GDK_Undo,		'&', '8' },
    { GDK_BackSpace,	'k', 'b' },
    { GDK_Insert,	'k', 'I' },
    { GDK_Delete,	'k', 'D' },
    { GDK_Home,		'k', 'h' },
    { GDK_End,		'@', '7' },
    { GDK_Prior,	'k', 'P' },
    { GDK_Next,		'k', 'N' },
    { GDK_Print,	'%', '9' },
    /* Keypad keys: */
    { GDK_KP_Left,	'k', 'l' },
    { GDK_KP_Right,	'k', 'r' },
    { GDK_KP_Up,	'k', 'u' },
    { GDK_KP_Down,	'k', 'd' },
    { GDK_KP_Insert,	'k', 'I' },
    { GDK_KP_Delete,	'k', 'D' },
    { GDK_KP_Home,	'k', 'h' },
    { GDK_KP_End,	'@', '7' },
    { GDK_KP_Prior,	'k', 'P' },
    { GDK_KP_Next,	'k', 'N' },
    /* End of list marker: */
    { 0, 0, 0 }
};




/*
 * This table holds all the X GUI command line options allowed.  This includes
 * the standard ones so that we can skip them when vim is started without the
 * GUI (but the GUI might start up later).
 * When changing this, also update doc/vim_gui.txt and the usage message!!!
 */
static XrmOptionDescRec cmdline_options[] =
{
    /* We handle these options ourselfs */
    {"-bg", ".background", XrmoptionSepArg, NULL},
    {"-background", ".background", XrmoptionSepArg, NULL},
    {"-fg", ".foreground", XrmoptionSepArg, NULL},
    {"-foreground", ".foreground", XrmoptionSepArg, NULL},
    {"-fn", ".font", XrmoptionSepArg, NULL},
    {"-font", ".font", XrmoptionSepArg, NULL},
    {"-boldfont", ".boldFont", XrmoptionSepArg, NULL},
    {"-italicfont", ".italicFont", XrmoptionSepArg, NULL},
    {"-geom", ".geometry", XrmoptionSepArg, NULL},
    {"-geometry", ".geometry", XrmoptionSepArg, NULL},
    {"-reverse", "*reverseVideo", XrmoptionNoArg, "True"},
    {"-rv", "*reverseVideo", XrmoptionNoArg, "True"},
    {"+reverse", "*reverseVideo", XrmoptionNoArg, "False"},
    {"+rv", "*reverseVideo", XrmoptionNoArg, "False"},
    {"-display", ".display", XrmoptionSepArg, NULL},
    {"-iconic", "*iconic", XrmoptionNoArg, "True"},
    {"-name", ".name", XrmoptionSepArg, NULL},
    {"-bw", ".borderWidth", XrmoptionSepArg, NULL},
    {"-borderwidth", ".borderWidth", XrmoptionSepArg, NULL},
    {"-xrm", NULL, XrmoptionResArg, NULL}
};

static int gui_argc = 0;
static char **gui_argv = NULL;

/*
 * Parse the GUI related command-line arguments.  Any arguments used are
 * deleted from argv, and *argc is decremented accordingly.  This is called
 * when vim is started, whether or not the GUI has been started.
 */
void
gui_mch_prepare(int *argc, char **argv)
{
    int arg;
    int i;

    /*
     * Move all the entries in argv which are relevant to X into gui_argv.
     */
    gui_argc = 0;
    gui_argv = (char **) lalloc((long_u) (*argc * sizeof(char *)), FALSE);
    if (gui_argv == NULL)
	return;

    gui_argv[gui_argc++] = argv[0];
    arg = 1;
    while (arg < *argc) {
	/* Look for argv[arg] in cmdline_options[] table */
	for (i = 0; i < XtNumber(cmdline_options); i++)
	    if (strcmp(argv[arg], cmdline_options[i].option) == 0)
		break;

	if (i < XtNumber(cmdline_options)) {
	    /* Remember finding "-rv" or "-reverse" */
	    if (strcmp("-rv", argv[arg]) == 0 ||
		strcmp("-reverse", argv[arg]) == 0) {
		found_reverse_arg = TRUE;
	    } else if ((strcmp("-fn", argv[arg]) == 0 ||
		      strcmp("-font", argv[arg]) == 0) &&
		      arg + 1 < *argc) {
		font_opt = argv[arg + 1];
	    }

	    /* Found match in table, so move it into gui_argv */
	    gui_argv[gui_argc++] = argv[arg];
	    if (--*argc > arg) {
		mch_memmove(&argv[arg], &argv[arg + 1],
			    (*argc - arg) * sizeof(char *));
		if (cmdline_options[i].argKind != XrmoptionNoArg) {
		    /* Move the options argument as well */
		    gui_argv[gui_argc++] = argv[arg];
		    if (--*argc > arg)
			mch_memmove(&argv[arg], &argv[arg + 1], (*argc - arg)
				    * sizeof(char *));
		}
	    }
	} else
	    ++arg;
    }
}


/*
 * This signal informs us about the need to rearrange our subwidgets.
 */
static gint
form_configure_event(GtkWidget * widget, GdkEventConfigure * event)
{
    gtk_form_freeze(GTK_FORM(form_window));
    gui_resize_window(event->width, event->height);
    gtk_form_thaw(GTK_FORM(form_window));

    return TRUE;
}

#ifdef GTK_HAVE_FEATURES_1_1_0
/*
 * This should be maybe compleatly removed.
 */
static gint
visibility_event(GtkWidget * widget, GdkEventVisibility * event)
{
    /* Just remember it for the creation of GCs before drawing
     * operations.
     */
    gui.visibility = event->state;

    return FALSE;
}
#endif

/*
 * This is a simple state machine:
 * BLINK_NONE   not blinking at all
 * BLINK_OFF    blinking, cursor is not shown
 * BLINK_ON     blinking, cursor is shown
 */

#define BLINK_NONE  0
#define BLINK_OFF   1
#define BLINK_ON    2

static int blink_state = BLINK_NONE;
static long_u blink_waittime = 700;
static long_u blink_ontime = 400;
static long_u blink_offtime = 250;
static guint blink_timer = 0;

void
gui_mch_set_blinking(long waittime, long on, long off)
{
    blink_waittime = waittime;
    blink_ontime = on;
    blink_offtime = off;
}

/*
 * Stop the cursor blinking.  Show the cursor if it wasn't shown.
 */
void
gui_mch_stop_blink()
{
    if (blink_timer) {
	gtk_timeout_remove(blink_timer);
	blink_timer = 0;
    }
    if (blink_state == BLINK_OFF)
	gui_update_cursor(TRUE, FALSE);
    blink_state = BLINK_NONE;
}

static gint
gui_gtk_blink_cb(gpointer data)
{
    if (blink_state == BLINK_ON) {
	gui_undraw_cursor();
	blink_state = BLINK_OFF;
	blink_timer = gtk_timeout_add(blink_offtime,
				   (GtkFunction) gui_gtk_blink_cb, NULL);
    } else {
	gui_update_cursor(TRUE, FALSE);
	blink_state = BLINK_ON;
	blink_timer = gtk_timeout_add(blink_ontime,
				   (GtkFunction) gui_gtk_blink_cb, NULL);
    }

    return FALSE;		/* don't happen again */
}

/*
 * Start the cursor blinking.  If it was already blinking, this restarts the
 * waiting time and shows the cursor.
 */
void
gui_mch_start_blink()
{
    if (blink_timer)
	gtk_timeout_remove(blink_timer);
    /* Only switch blinking on if none of the times is zero */
    if (blink_waittime && blink_ontime && blink_offtime && gui.in_focus) {
	blink_timer = gtk_timeout_add(blink_waittime,
				   (GtkFunction) gui_gtk_blink_cb, NULL);
	blink_state = BLINK_ON;
	gui_update_cursor(TRUE, FALSE);
    }
}

static void
enter_notify_event(GtkContainer * container, gpointer data)
{
    if (blink_state == BLINK_NONE)
	gui_mch_start_blink();

    /* make sure keybord input goes there */
    gtk_widget_grab_focus(drawing_area);
}

static void
leave_notify_event(GtkContainer * container, gpointer data)
{
    if (blink_state != BLINK_NONE)
	gui_mch_stop_blink();
}

static void
focus_in_event(GtkContainer * container, gpointer data)
{
    gui_focus_change(TRUE);

    if (blink_state == BLINK_NONE)
	gui_mch_start_blink();

    /* make sure keybord input goes there */
    gtk_widget_grab_focus(drawing_area);
}

static void
focus_out_event(GtkContainer * container, gpointer data)
{
    gui_focus_change(FALSE);

    if (blink_state != BLINK_NONE)
	gui_mch_stop_blink();
}

static gint
selection_clear_event(GtkWidget * widget, GdkEventSelection * event)
{
    clip_lose_selection();
    if (gtk_main_level() > 0)
	gtk_main_quit();
    return TRUE;
}

static void
selection_received_event(GtkWidget * widget, GtkSelectionData * data)
{
    int motion_type;
    long_u len;
    char_u *p;

    if ((!data->data) || (data->length <= 0)) {
	/* clip_free_selection(); ??? */
	if (gtk_main_level() > 0)
	    gtk_main_quit();

	return;
    }

    motion_type = MCHAR;
    p = (char_u *) data->data;
    len = data->length;

    if (data->type == clipboard.atom) {
	motion_type = *p++;
	len--;
    }
    clip_yank_selection(motion_type, p, (long) len);
    if (gtk_main_level() > 0)
	gtk_main_quit();
}

#ifdef GTK_HAVE_FEATURES_1_1_0
/*
 * Prepare our selection data for passing it to the external selection
 * client.
 */
static void
selection_get_event(GtkWidget *widget,
		    GtkSelectionData *selection_data,
		    guint      info,
		    guint      time,
		    gpointer   data)
{
    char_u *string;
    char_u *result;
    long_u length;
    int motion_type;
    GdkAtom type = GDK_NONE;

    if (!clipboard.owned)
	return;			/* Shouldn't ever happen */

    if (info != SELECTION_STRING && info != SELECTION_CLIPBOARD)
	return;

    clip_get_selection();
    length = selection_data->length;

    motion_type = clip_convert_selection(&string, (long_u *) & length);
    if (motion_type < 0)
	return;

    /* For our own format, the first byte contains the motion type */
    if (info == SELECTION_CLIPBOARD)
	length++;

    if ((result = (char_u *) malloc(2 * length)) == NULL) {
	vim_free(string);
	return;
    }

    switch (info) {
	case SELECTION_STRING:
	    mch_memmove(result, string, (size_t) length);
	    type = GDK_TARGET_STRING;
	    break;
	default :
	    result[0] = motion_type;
	    mch_memmove(result + 1, string, (size_t) (length - 1));
	    type = clipboard.atom;
	    break;
    }
    selection_data->type = selection_data->target;
    selection_data->format = 8;	/* 8 bits per char */

    gtk_selection_data_set(selection_data, type, 8, result, length);
    vim_free(string);
    vim_free(result);
}

#else	/* not GTK_HAVE_FEATURES_1_1_0 */

static void
selection_handler(GtkWidget *widget, GtkSelectionData *selection_data,
		  gpointer data)
{
    char_u *string;
    char_u *result;
    long_u length;
    int motion_type;

    if (!clipboard.owned)
	return;                 /* Shouldn't ever happen */

    if (selection_data->target != clipboard.atom &&
	selection_data->target != GDK_TARGET_STRING)
    {
	return;
    }

    clip_get_selection();
    length = selection_data->length;

    motion_type = clip_convert_selection(&string, (long_u *) & length);
    if (motion_type < 0)
	return;

    /* For our own format, the first byte contains the motion type */
    if (selection_data->target == clipboard.atom)
	length++;

    result = (char_u *) malloc(2 * length);
    if (result == NULL) {
	vim_free(string);
	return;
    }

    if (selection_data->target == GDK_TARGET_STRING)
	mch_memmove(result, string, (size_t)length);
    else {
	result[0] = motion_type;
	mch_memmove(result + 1, string, (size_t)(length - 1));
    }

    selection_data->type = selection_data->target;
    selection_data->format = 8; /* 8 bits per char */

    gtk_selection_data_set(selection_data, selection_data->target, 8,
	    result, length);
    vim_free(string);
    vim_free(result);
}

#endif	/* GTK_HAVE_FEATURES_1_1_0 */


/*
 * Initialise the X GUI.  Create all the windows, set up all the call-backs etc.
 * Returns OK for success, FAIL when the GUI can't be started.
 */
int
gui_mch_init()
{
    gtk_init(&gui_argc, &gui_argv);

    vim_free(gui_argv);

    if (!GDK_DISPLAY()) {
	gui.dying = TRUE;
	EMSG("cannot open display");
	return FAIL;
    }

    /* Uncomment this to enable synchronous mode for debugging */
    /* XSynchronize(gui.dpy, True); */

    /* Initialize values */
    gui.rev_video = FALSE;
    gui.border_width = 2;
    gui.scrollbar_width = SB_DEFAULT_WIDTH;

    gui.scrollbar_height = gui.scrollbar_width;
    /* Don't change the menu height values used in gui.c at runtime */
    gui.menu_height_fixed = TRUE;

    /* Set default foreground and background colours. */
    gui.norm_pixel = gui.def_norm_pixel;
    gui.back_pixel = gui.def_back_pixel;

    main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_container_border_width(GTK_CONTAINER(main_window), 0);
    gtk_widget_set_events(main_window, GDK_VISIBILITY_NOTIFY_MASK);

    /* We need to realize the main_window this early, since otherwise
     * the toolbar pixmaps will not have a window to which they could be
     * attached.
     */
    gtk_widget_realize(main_window);
    /* Add an icon to the main window. For fun and convenience of the user */
    if (vim_strchr(p_guioptions, GO_ICON) != NULL) {
#include "../runtime/vim32x32.xpm"
	static GdkPixmap *icon = NULL;
	static GdkBitmap *icon_mask = NULL;

	if (!icon) {
	    icon = gdk_pixmap_create_from_xpm_d(main_window->window,
						&icon_mask,
						NULL,
						magick);
	}
	gdk_window_set_icon(main_window->window, NULL, icon, icon_mask);
    }


#ifdef GTK_HAVE_FEATURES_1_1_0
    /* FIXME: this should eventually get the accelgroup of the main_window */
    accel_group = gtk_accel_group_get_default();
#endif

    main_vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(main_window), main_vbox);
    gtk_widget_show(main_vbox);

    /*
     * I have just removed the attach/dettach handle, since:
     * 1. It's buggy like hell in GTK+ itself and we don't like
     *    the users to blame us for others errors.
     * 2. No single sane GUI stnadard uses them (Netscape!).
     *    We can however consider it later for the toolbar.
     * 3. The functionality of it is just a gimmic.
     * 4. It's just causing more trouble than it's worth...
     */
    menu_bar = gtk_menu_bar_new();
    gtk_box_pack_start(GTK_BOX(main_vbox), menu_bar, FALSE, TRUE, 0);

    /* Create the tool-bar widget. There is no need to create
     * one-shoot functions. They will only disturb code reading.
     */
    if (p_toolbar == 2)
	tool_bar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
    else if (p_toolbar == 3)
	tool_bar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_TEXT);
    else
	tool_bar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);

#ifdef GTK_HAVE_FEATURES_1_1_0
    gtk_toolbar_set_button_relief(GTK_TOOLBAR(tool_bar), GTK_RELIEF_NONE);
#endif
    /* esthetics nit picking */
    gtk_toolbar_append_space(GTK_TOOLBAR(tool_bar));
    gtk_box_pack_start(GTK_BOX(main_vbox), tool_bar, FALSE, TRUE, 0);

    form_window = gtk_form_new();
    gtk_container_border_width(GTK_CONTAINER(form_window), 0);
    gtk_widget_set_events(form_window, GDK_EXPOSURE_MASK);

    drawing_area = gtk_drawing_area_new();
    /* Determine which event we will filter. */
    gtk_widget_set_events(drawing_area,
			  GDK_EXPOSURE_MASK |
			  GDK_ENTER_NOTIFY_MASK |
			  GDK_LEAVE_NOTIFY_MASK |
			  GDK_BUTTON_PRESS_MASK |
			  GDK_BUTTON_RELEASE_MASK |
			  GDK_POINTER_MOTION_MASK |
			  GDK_POINTER_MOTION_HINT_MASK);
    gtk_widget_show(drawing_area);
    gtk_form_put(GTK_FORM(form_window), drawing_area, 0, 0);
    gtk_widget_show(form_window);
    gtk_box_pack_start(GTK_BOX(main_vbox), form_window, TRUE, TRUE, 0);
    if (0) {
	GtkWidget *tmp;
	tmp = gtk_menu_bar_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), tmp, FALSE, TRUE, 0);

	gtk_widget_show(tmp);
    }

    gtk_signal_connect(GTK_OBJECT(main_window), "key_press_event",
		       (GtkSignalFunc) key_press_event, NULL);

    /* FIXME: duno what this is all about */
    /* Check if reverse video needs to be applied (on Sun it's done by X) */
    if (gui.rev_video && gui_mch_get_lightness(gui.back_pixel)
	> gui_mch_get_lightness(gui.norm_pixel)) {
	gui.norm_pixel = gui.def_back_pixel;
	gui.back_pixel = gui.def_norm_pixel;
	gui.def_norm_pixel = gui.norm_pixel;
	gui.def_back_pixel = gui.back_pixel;
    }
    gui.visibility = GDK_VISIBILITY_UNOBSCURED;
    clipboard.atom = gdk_atom_intern("VIM_SELECTION", FALSE);

    /*
     * Start out by adding the configured border width into the border offset
     */
    gui.border_offset = gui.border_width;

#ifdef GTK_HAVE_FEATURES_1_1_0
    gtk_signal_connect(GTK_OBJECT(main_window), "visibility_notify_event",
		       GTK_SIGNAL_FUNC(visibility_event), NULL);
#endif
    gtk_signal_connect(GTK_OBJECT(drawing_area), "expose_event",
		       GTK_SIGNAL_FUNC(expose_event), NULL);

    /*
     * Notify the fixed area about the need to resize the contents of the
     * form_window, which we use for random positioning of the included
     * components.
     */
    gtk_signal_connect(GTK_OBJECT(form_window), "configure_event",
		       GTK_SIGNAL_FUNC(form_configure_event), NULL);

    /*
     * Only install these enter/leave callbacks when 'p' in 'guioptions'.
     * Only needed for some window managers.
     */
    if (vim_strchr(p_guioptions, GO_POINTER) != NULL) {
	gtk_signal_connect(GTK_OBJECT(drawing_area), "leave_notify_event",
			   GTK_SIGNAL_FUNC(leave_notify_event), NULL);
	gtk_signal_connect(GTK_OBJECT(drawing_area), "enter_notify_event",
			   GTK_SIGNAL_FUNC(enter_notify_event), NULL);
    }

    gtk_signal_connect(GTK_OBJECT(main_window), "focus_out_event",
			   GTK_SIGNAL_FUNC(focus_out_event), NULL);
    gtk_signal_connect(GTK_OBJECT(main_window), "focus_in_event",
			   GTK_SIGNAL_FUNC(focus_in_event), NULL);

    gtk_signal_connect(GTK_OBJECT(drawing_area), "motion_notify_event",
		       GTK_SIGNAL_FUNC(motion_notify_event), NULL);
    gtk_signal_connect(GTK_OBJECT(drawing_area), "button_press_event",
		       GTK_SIGNAL_FUNC(button_press_event), NULL);
    gtk_signal_connect(GTK_OBJECT(drawing_area), "button_release_event",
		       GTK_SIGNAL_FUNC(button_release_event), NULL);
    /*
     * Add selection handler functions.
     */

    gtk_signal_connect(GTK_OBJECT(drawing_area), "selection_clear_event",
		       GTK_SIGNAL_FUNC(selection_clear_event), NULL);
    gtk_signal_connect(GTK_OBJECT(drawing_area), "selection_received",
		       GTK_SIGNAL_FUNC(selection_received_event), NULL);

#ifdef GTK_HAVE_FEATURES_1_1_0
gtk_selection_add_target(drawing_area, GDK_SELECTION_PRIMARY,
			    GDK_TARGET_STRING, SELECTION_STRING);
    gtk_selection_add_target(drawing_area, GDK_SELECTION_PRIMARY,
			    clipboard.atom, SELECTION_CLIPBOARD);
    gtk_signal_connect(GTK_OBJECT(drawing_area), "selection_get",
		       GTK_SIGNAL_FUNC(selection_get_event), NULL);
#else
    gtk_selection_add_handler(drawing_area, GDK_SELECTION_PRIMARY,
			      GDK_TARGET_STRING, selection_handler, NULL);
    gtk_selection_add_handler(drawing_area, GDK_SELECTION_PRIMARY,
			      clipboard.atom, selection_handler, NULL);
#endif
    return OK;
}


/*
 * Called when the foreground or background color has been changed.
 */
void
gui_mch_new_colors()
{

    /* This used to change the graphics contexts directly but we are currently
     * manipulating them where desired.
     */
    if (drawing_area && drawing_area->window) {
	GdkColor color;
	color.pixel = gui.back_pixel;
	gdk_window_set_background(drawing_area->window, &color);
    }
}


static int
get_base_width(void)
{
    int width;
    GtkWidget *scroll_bar;

    width =  gui.border_offset * 2;

    if (gui.which_scrollbars[SBAR_LEFT]) {
	scroll_bar = firstwin->w_scrollbars[SBAR_LEFT].id;
	if (scroll_bar && GTK_WIDGET_REALIZED(scroll_bar)
		&& GTK_WIDGET_VISIBLE(scroll_bar))
	    width += scroll_bar->allocation.width;
    }

    if (gui.which_scrollbars[SBAR_RIGHT]) {
	scroll_bar = firstwin->w_scrollbars[SBAR_RIGHT].id;
	if (scroll_bar && GTK_WIDGET_REALIZED(scroll_bar)
		&& GTK_WIDGET_VISIBLE(scroll_bar))
	    width += scroll_bar->allocation.width;
    }

    return width;
}

static int
get_base_height(void)
{
    int height;

    height = gui.border_offset * 2;
    if (gui.which_scrollbars[SBAR_BOTTOM]) {
	GtkWidget *scroll_bar;

	scroll_bar = gui.bottom_sbar.id;
	if (scroll_bar && GTK_WIDGET_REALIZED(scroll_bar)
		&& GTK_WIDGET_VISIBLE(scroll_bar))
	    height += scroll_bar->allocation.height;
    }
    if (tool_bar && GTK_WIDGET_REALIZED(tool_bar)
	    && GTK_WIDGET_VISIBLE(tool_bar))
	height += tool_bar->allocation.height;
    if (menu_bar && GTK_WIDGET_REALIZED(menu_bar)
	    && GTK_WIDGET_VISIBLE(menu_bar))
	height += menu_bar->allocation.height;

    return height;
}

/*
 * Argh!!! Once again we need to deal with an ommission in GTK+ by
 * resorting to direct Xlib calls. Fortunatly I know how to do it :-).
 */
static void
update_window_manager_hints(int width, int height)
{
    XSizeHints size_hints;

    gdk_window_resize(main_window->window,
		      get_base_width() + width,
		      get_base_height() + height);

    /* Don't bother until the main window is really there */
    if (!GTK_WIDGET_REALIZED(main_window))
	return;

    size_hints.flags = PResizeInc;
    size_hints.width_inc = gui.char_width;
    size_hints.height_inc = gui.char_height;
    size_hints.flags |= PBaseSize;
    size_hints.base_width = get_base_width();
    size_hints.base_height = get_base_height();

    XSetWMNormalHints(GDK_DISPLAY(),
	    GDK_WINDOW_XWINDOW(main_window->window),
	    &size_hints);
}

/*
 * Open the GUI window which was created by a call to gui_mch_init().
 */
int
gui_mch_open()
{
    GdkColor color;

    gtk_widget_show(main_window);

    /* This will make us wait until infally all widgets arrived where they
     * should be.
     */
    gtk_main();

    /* We resize the main window only actually after it had been realized.
     * After that the whole size control is taked over by the event system.
     *
     * FIXME: take different desired startup sizes into account too.
     * Currently Sun's terminal sizes are just fine with me.
     */
    update_window_manager_hints(gui.char_width * 80, gui.char_height * 35);

    /* Now since the main window is really visible, we will setup the drawing
     * colors, which need a real window for working.  GTK widget colors are
     * already all fine, so there is no special way to configure them
     * provided.
     */

    /* Get the colors from the "Normal" and "Menu" group (set in syntax.c or
     * in a vimrc file)
     */
    set_normal_colors();

    /* Check that none of the colors are the same as the background color */
    gui_check_colors();

    /* Get the colors for the highlight groups (gui_check_colors() might have
     * changed them).
     */
    highlight_gui_started();	/* re-init colors and fonts */

    color.pixel = gui.back_pixel;
    gdk_window_set_background(drawing_area->window, &color);

    /* Create a mouse pointer that is blank.  FIXME: This doesn't seem to work
     * properly. However I didn't investigate it at all, since I'm not using
     * the corresponding option for myself.
     */
    {
	Pixmap blank_pixmap = XCreatePixmap(GDK_DISPLAY(),
		GDK_WINDOW_XWINDOW(drawing_area->window),
		1, 1, 1);
	GC gc = XCreateGC(GDK_DISPLAY(), blank_pixmap, (unsigned long)0, (XGCValues *)0);
	XDrawPoint(GDK_DISPLAY(), blank_pixmap, gc, 0, 0);
	XFreeGC(GDK_DISPLAY(), gc);
	gui. blank_pointer = XCreatePixmapCursor(GDK_DISPLAY(), blank_pixmap, blank_pixmap,
		(XColor *)(&gui.norm_pixel),
		(XColor *)(&gui.norm_pixel), 0, 0);
    }

    /*
     * FIXME: needed!!!!
     * Add a callback for the save-yourself event.
     */
    gtk_signal_connect(GTK_OBJECT(main_window), "destroy",
		       GTK_SIGNAL_FUNC(destroy_callback), NULL);

    return OK;
}


void
gui_mch_exit(int rc)
{
    gtk_exit(0);
}

/*
 * This function get's called only once at startup. Thereafter we take over the
 * whole control of sizes by the event mechanism of the GUI.
 */
void
gui_mch_set_winsize(int width, int height,
		    int min_width, int min_height,
		    int base_width, int base_height)
{
    gtk_widget_set_usize(main_window, min_width, min_height);
    gtk_form_set_size(GTK_FORM(form_window), width, height);

    /* Make sure all the events generated here will take immediatly
     * effect before anything else will try to reuse the size values.
     */
    gtk_main_iteration_do(FALSE);
}



/*
 * Those parameters are used to make sure the initial window doesn't get bigger
 * then the main window. However this doesn't take the Window Manager
 * decorations into account currently.
 */
void
gui_mch_get_screen_dimensions(int *screen_w, int *screen_h)
{
    *screen_w = gdk_screen_width();
    *screen_h = gdk_screen_height();
}

void
gui_mch_enable_menu(int flag)
{
    int width;
    int height;

    height = drawing_area->allocation.height - 2 * gui.border_offset;
    width = drawing_area->allocation.width - 2 * gui.border_offset;

    if (flag)
	gtk_widget_show(menu_bar);
    else
	gtk_widget_hide(menu_bar);

    update_window_manager_hints(width, height);
}


void
gui_mch_show_toolbar(int showit)
{
    if (tool_bar == NULL)
	return;

    if (p_toolbar == 0 || !showit) {
	if (GTK_WIDGET_VISIBLE(tool_bar)) {
	    int width;
	    int height;

	    height = drawing_area->allocation.height - 2 * gui.border_offset;
	    width = drawing_area->allocation.width - 2 * gui.border_offset;

	    gtk_widget_hide(tool_bar);

	    update_window_manager_hints(width, height);
	}
    } else {
	if (p_toolbar == 2) {
	    gtk_toolbar_set_tooltips(GTK_TOOLBAR(tool_bar), FALSE);
	    gtk_toolbar_set_style(GTK_TOOLBAR(tool_bar), GTK_TOOLBAR_TEXT);
	} else if (p_toolbar == 3) {
	    gtk_toolbar_set_tooltips(GTK_TOOLBAR(tool_bar), FALSE);
	    gtk_toolbar_set_style(GTK_TOOLBAR(tool_bar), GTK_TOOLBAR_BOTH);
	} else {
	    gtk_toolbar_set_tooltips(GTK_TOOLBAR(tool_bar), TRUE);
	    gtk_toolbar_set_style(GTK_TOOLBAR(tool_bar), GTK_TOOLBAR_ICONS);
	}

	/*
	 * Black on white will catch the attention of the user more likely then
	 * black on grey. However due to errors / omissions in GTK+-1.1.5 this
	 * cased to work properly. I had been looking after it and there seems
	 * to be no easy fix for it there. (--mdcki)
	 */
	gtk_tooltips_set_colors(GTK_TOOLBAR(tool_bar)->tooltips,
				&tool_bar->style->white,
				&tool_bar->style->black);
	if (!GTK_WIDGET_VISIBLE(tool_bar)) {
	    int width;
	    int height;

	    height = drawing_area->allocation.height - 2 * gui.border_offset;
	    width = drawing_area->allocation.width - 2 * gui.border_offset;

	    gtk_widget_show(tool_bar);

	    update_window_manager_hints(width, height);
	}
    }
    gtk_main_iteration_do(FALSE);
}


/*
 * Initialise vim to use the font with the given name.
 * Return FAIL if the font could not be loaded, OK otherwise.
 */
int
gui_mch_init_font(char_u * font_name)
{
    GdkFont *font = NULL;
    char *chunk[14], *dup, *tmp;
    int len, i;

    /*
     * If none of the fonts in 'font' could be loaded, try the default, which
     * should be present on all X11 servers.
     */
    if (font_name == NULL)
	font_name = (char_u *) DFLT_FONT;

    if (STRCMP(font_name, "*") == 0) {
	/*
	 * Request for a font handling dialog.
	 * Not quite sure we should handle this here...
	 */

#ifdef GTK_HAVE_FEATURES_1_1_0
	/*
	 * NOTE about font selection widget: this can easily be backported to
	 * gtk-1.0.x.
	 */
	GtkWidget *window = NULL;

	window = gtk_font_selection_dialog_new("Font Selection Dialog");
	gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(window), "destroy",
		GTK_SIGNAL_FUNC(font_selection_destroy),
		&window);
	gtk_signal_connect(
		GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(window)->ok_button),
		"clicked", GTK_SIGNAL_FUNC(font_selection_ok),
		GTK_FONT_SELECTION_DIALOG(window));
	gtk_signal_connect_object(
		GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(window)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
		GTK_OBJECT(window));
	gtk_widget_show(window);
	gtk_main();

	font_name = selected_font_name;
#endif
    }
    if (font == NULL)
	font = gui_mch_get_font(font_name, FALSE);

    if (font == NULL)
	return FAIL;

    if (gui.norm_font != 0)
	gdk_font_unref(font);
    gui.norm_font = font;
    gui.char_width = ((XFontStruct *) GDK_FONT_XFONT(font))->max_bounds.width;
    gui.char_height = font->ascent + font->descent;
    gui.char_ascent = font->ascent;

    /* Set the fontname, which will be used for information purposes */
    hl_set_font_name(font_name);

    /* There is only one excuse I can give for the following attempt
     * to manage font styles:
     *
     * I HATE THE BRAIN DEAD WAY X11 IS HANDLING FONTS (--mdcki)
     */
    if ((dup = g_strdup((const char *)font_name)) == NULL)
	return FAIL;

    /* slipt up the whole */
    i = 0;
    for (tmp = dup; *tmp != '\0'; ++tmp) {
	if (*tmp == '-') {
	    *tmp = '\0';
	    chunk[i] = tmp + 1;
	    ++i;
	}
    }
    vim_free(dup);

    if (i == 14) {
	char *bold_name = NULL;
	char *ital_name = NULL;
	char *italbold_name = NULL;

	/* font name was compleate */
	len = strlen((const char *)font_name) + 32;
	bold_name = (char *)alloc(len);
	ital_name = (char *)alloc(len);
	italbold_name = (char *)alloc(len);
	if (bold_name == NULL || ital_name == NULL || italbold_name == NULL) {
	    vim_free(bold_name);
	    vim_free(ital_name);
	    vim_free(italbold_name);
	    return FAIL;
	}

	*bold_name = '\0';
	*ital_name = '\0';
	*italbold_name = '\0';

	for (i = 0; i < 14; ++i) {
	    strcat(bold_name, "-");
	    strcat(ital_name, "-");
	    strcat(italbold_name, "-");
	    strcat(bold_name, (i != 2) ? chunk[i] : "bold");
	    strcat(ital_name, (i != 3) ? chunk[i] : "o");

	    if (i != 2 && i != 3)
		strcat(italbold_name, chunk[i]);
	    else {
		if (i == 2)
		    strcat(italbold_name, "bold");
		else if (i == 3)
		    strcat(italbold_name, "o");
	    }
	}

	font = gui_mch_get_font((char_u *)bold_name, FALSE);
	if (font != NULL)
	    gui.bold_font = font;
	else if (gui.bold_font) {
	    gdk_font_unref(gui.bold_font);
	    gui.bold_font = NULL;
	}

	font = gui_mch_get_font((char_u *)ital_name, FALSE);
	if (font != NULL)
	    gui.ital_font = font;
	else if (gui.ital_font) {
	    gdk_font_unref(gui.ital_font);
	    gui.ital_font = NULL;
	}

	font = gui_mch_get_font((char_u *)italbold_name, FALSE);
	if (font != NULL)
	    gui.boldital_font = font;
	else if (gui.boldital_font)     {
	    gdk_font_unref(gui.boldital_font);
	    gui.boldital_font = NULL;
	}

	vim_free(bold_name);
	vim_free(ital_name);
	vim_free(italbold_name);
    }

    /* Synchronize the fonts used in user input dialogs, since otherwise
     * search/replace will be esp annoyig in case of international font usage.
     */
    gui_gtk_synch_fonts();

    /* Preserve the logical dimensions of the screen.  However for some reasons
     * this doesn't work quite perfectly.
     */
    update_window_manager_hints(gui.num_cols * gui.char_width,
	    gui.num_rows * gui.char_height);

    return OK;
}


GuiFont
gui_mch_get_font(char_u * name, int giveErrorIfMissing)
{
    GdkFont *font;

    if (!gui.in_use)		/* can't do this when GUI not running */
	return (GuiFont) 0;

    font = gdk_font_load((const gchar *) name);

    if (font == NULL) {
	if (giveErrorIfMissing)
	    EMSG2("Unknown font: %s", name);
	return (GuiFont) 0;
    }

    /* reference this font as beeing in use */
    gdk_font_ref(font);
    if (((XFontStruct *) GDK_FONT_XFONT(font))->max_bounds.width
	!= ((XFontStruct *) GDK_FONT_XFONT(font))->min_bounds.width) {
	EMSG2("Font \"%s\" is not fixed-width", name);
	gdk_font_unref(font);

	return (GuiFont) 0;
    }
    return (GuiFont) font;
}

/*
 * Set the current text font.
 *
 * We don't do it in the graphics context but use simple this global variable
 * instead. However this should be moved to the gui global structure with all
 * the other statics together some time...
 *
 * This will be necessary when we go finally hunting for multiple main window
 * instances.
 */

static GdkFont *current_font = NULL;

void
gui_mch_set_font(GuiFont font)
{
    current_font = (GdkFont *) font;
}

/*
 * Return TRUE if the two fonts given are equivalent.
 */
int
gui_mch_same_font(GuiFont f1, GuiFont f2)
{
    return gdk_font_equal((GdkFont *) f1, (GdkFont *) f2);
}

/*
 * If a font is not going to be used, free its structure.
 */
void
gui_mch_free_font(GuiFont font)
{
    if (font)
	gdk_font_unref((GdkFont *) font);
}


/*
 * Return the Pixel value (color) for the given color name.  This routine was
 * pretty much taken from example code in the Silicon Graphics OSF/Motif
 * Programmer's Guide.
 * Return -1 for error.
 */
GuiColor
gui_mch_get_color(char_u * name)
{
    int i;
    static char *(vimnames[][2]) =
    {
    /* A number of colors that some X11 systems don't have */
	{"LightRed", "#FFA0A0"},
	{"LightGreen", "#80FF80"},
	{"LightMagenta", "#FFA0FF"},
	{"DarkCyan", "#008080"},
	{"DarkBlue", "#0000c0"},
	{"DarkRed", "#c00000"},
	{"DarkMagenta", "#c000c0"},
	{"DarkGrey", "#c0c0c0"},
	{NULL, NULL}
    };

    if (!gui.in_use)		/* can't do this when GUI not running */
	return (GuiColor)(-1);

    while (name != NULL) {
	GdkColor color;

	if (gdk_color_parse((const gchar *)name, &color)) {
	    GdkColormap *colormap;
	    colormap = gdk_window_get_colormap(drawing_area->window);
	    gdk_color_alloc(colormap, &color);

	    return (GuiColor) color.pixel;
	}
	/* add a few builtin names */
	for (i = 0;; ++i) {
	    if (vimnames[i][0] == NULL) {
		name = NULL;
		break;
	    }
	    if (STRICMP(name, vimnames[i][0]) == 0) {
		name = (char_u *) vimnames[i][1];
		break;
	    }
	}
    }

    return (GuiColor)(-1);
}

/*
 * Set the current text foreground color.
 */
void
gui_mch_set_fg_color(GuiColor color)
{
    foreground_color.pixel = (Pixel) color;
}

/*
 * Set the current text background color.
 */
void
gui_mch_set_bg_color(GuiColor color)
{
    background_color.pixel = (Pixel) color;
}

/*
 * Use the blank mouse pointer or not.
 *
 * hide: TRUE = use blank ptr, FALSE = use parent ptr
 */
void
gui_mch_mousehide(int hide)
{
    if (gui.pointer_hidden != hide) {
	if (hide)
	    XDefineCursor(GDK_DISPLAY(),
		    GDK_WINDOW_XWINDOW(drawing_area->window),
		    gui.blank_pointer);
	else
	    XUndefineCursor(GDK_DISPLAY(),
		    GDK_WINDOW_XWINDOW(drawing_area->window));
	gui.pointer_hidden = hide;
    }
}

void
gui_mch_draw_string(int row, int col, char_u * s, int len, int flags)
{
    GdkGC *gc;

    if (!current_font)
	return;

    gc = gdk_gc_new(drawing_area->window);
    gdk_gc_set_exposures(gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);

    if (flags & DRAW_TRANSP) {
	gdk_gc_set_foreground(gc, &foreground_color);
	gdk_draw_text(drawing_area->window,
		      current_font,
		      gc,
		      TEXT_X(col), TEXT_Y(row),
		      (const gchar *)s, len);
    } else {
	int width;
	int height;
	width = gdk_text_width(current_font, (const gchar *)s, len);
	height = gui.char_height;

	gdk_gc_set_foreground(gc, &background_color);
	gdk_draw_rectangle(drawing_area->window,
			   gc,
			   TRUE,
			   FILL_X(col), FILL_Y(row), width, height);
	gdk_gc_set_foreground(gc, &foreground_color);
	gdk_draw_text(drawing_area->window,
		      current_font,
		      gc,
		      TEXT_X(col), TEXT_Y(row),
		      (const gchar *)s, len);
    }

    /* redraw the contents with an offset of 1 to emulate bold */
    if (flags & DRAW_BOLD) {
	gdk_draw_text(drawing_area->window,
		      current_font,
		      gc,
		      TEXT_X(col) + 1, TEXT_Y(row),
		      (const gchar *)s, len);
    }

    if (flags & DRAW_UNDERL) {
	gdk_draw_line(drawing_area->window,
		      gc, FILL_X(col),
	FILL_Y(row + 1) - 1, FILL_X(col + len) - 1, FILL_Y(row + 1) - 1);
    }
    gdk_gc_destroy(gc);
}

/*
 * Return OK if the key with the termcap name "name" is supported.
 */
int
gui_mch_haskey(char_u * name)
{
    int i;

    for (i = 0; special_keys[i].key_sym != 0; i++)
	if (name[0] == special_keys[i].code0 &&
	    name[1] == special_keys[i].code1)
	    return OK;
    return FAIL;
}

/*
 * Return the text window-id and display.  Only required for X-based GUI's
 */
int
gui_get_x11_windis(Window * win, Display ** dis)
{
    /*
     * This is currently only used by the code which is handling the
     * window manager title of this program.
     */
    *win = GDK_WINDOW_XWINDOW(main_window->window);
    *dis = GDK_DISPLAY();

    return OK;
}

void
gui_mch_beep()
{
    gdk_beep();
}

void
gui_mch_flash()
{
    GdkGCValues values;
    GdkGC *invert_gc;
    GdkColor foreground;
    GdkColor background;

    foreground.pixel = gui.norm_pixel ^ gui.back_pixel;
    background.pixel = gui.norm_pixel ^ gui.back_pixel;

    values.foreground = foreground;
    values.background = background;
    values.function = GDK_XOR;
    invert_gc = gdk_gc_new_with_values(drawing_area->window,
				       &values,
				       GDK_GC_FOREGROUND |
				       GDK_GC_BACKGROUND |
				       GDK_GC_FUNCTION);
    gdk_gc_set_exposures(invert_gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);

    /* Do a visual beep by changing back and forth in some undetermined way,
     * the foreground and background colors.  This is due to the fact that
     * there can't be really any prediction about the effects of XOR on
     * arbitrary X11 servers. However this seems to be enought for what we
     * intend it to do.
     */
    gdk_draw_rectangle(drawing_area->window, invert_gc,
		       TRUE,
		       0, 0,
		       FILL_X((int) Columns) + gui.border_offset,
		       FILL_Y((int) Rows) + gui.border_offset);

    gdk_flush();
    ui_delay(20L, TRUE);	/* wait 1/50 of a second */
    gdk_draw_rectangle(drawing_area->window, invert_gc,
		       TRUE,
		       0, 0,
		       FILL_X((int) Columns) + gui.border_offset,
		       FILL_Y((int) Rows) + gui.border_offset);

    gdk_gc_destroy(invert_gc);
}

/*
 * Invert a rectangle from row r, column c, for nr rows and nc columns.
 */
void
gui_mch_invert_rectangle(int r, int c, int nr, int nc)
{
    GdkGCValues values;
    GdkGC *invert_gc;
    GdkColor foreground;
    GdkColor background;

    foreground.pixel = gui.norm_pixel ^ gui.back_pixel;
    background.pixel = gui.norm_pixel ^ gui.back_pixel;

    values.foreground = foreground;
    values.background = background;
    values.function = GDK_XOR;
    invert_gc = gdk_gc_new_with_values(drawing_area->window,
				       &values,
				       GDK_GC_FOREGROUND |
				       GDK_GC_BACKGROUND |
				       GDK_GC_FUNCTION);
    gdk_gc_set_exposures(invert_gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);
    gdk_draw_rectangle(drawing_area->window, invert_gc,
		       TRUE,
		       FILL_X(c), FILL_Y(r),
		       (nc) * gui.char_width, (nr) * gui.char_height);
    gdk_gc_destroy(invert_gc);
}

/*
 * Iconify the GUI window.
 */
void
gui_mch_iconify()
{
    XIconifyWindow(GDK_DISPLAY(),
		   GDK_WINDOW_XWINDOW(main_window->window),
		   DefaultScreen(GDK_DISPLAY()));
}

/*
 * Draw a cursor without focus.
 */
void
gui_mch_draw_hollow_cursor(GuiColor color)
{
    GdkGC *gc;
    gui_mch_set_fg_color(color);

    gc = gdk_gc_new(drawing_area->window);
    gdk_gc_set_exposures(gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);
    gdk_gc_set_foreground(gc, &foreground_color);
    gdk_draw_rectangle(drawing_area->window, gc,
		       FALSE,
		       FILL_X(gui.col), FILL_Y(gui.row),
		       gui.char_width - 1, gui.char_height - 1);
    gdk_gc_destroy(gc);
}

/*
 * Draw part of a cursor, "w" pixels wide, and "h" pixels high, using
 * color "color".
 */
void
gui_mch_draw_part_cursor(int w, int h, GuiColor color)
{
    GdkGC *gc;
    gui_mch_set_fg_color(color);

    gc = gdk_gc_new(drawing_area->window);
    gdk_gc_set_exposures(gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);
    gdk_gc_set_foreground(gc, &foreground_color);
    gdk_draw_rectangle(drawing_area->window, gc,
		       TRUE,
#ifdef RIGHTLEFT
    /* vertical line should be on the right of current point */
	   State != CMDLINE && curwin->w_p_rl ? FILL_X(gui.col + 1) - w :
#endif
		       FILL_X(gui.col),
		       FILL_Y(gui.row) + gui.char_height - h,
		       w, h);
    gdk_gc_destroy(gc);
}


/*
 * This idle function is really idle since it's only purpose is to
 * cach up with s single X11 event.
 */
static gint
idle_function(GtkWidget * label)
{
    if (gtk_main_level() > 0)
	gtk_main_quit();

    return FALSE;		/* don't happen again */
}

/*
 * Catch up with any queued X11 events.  This may put keyboard input into the
 * input buffer, call resize call-backs, trigger timers etc.  If there is
 * nothing in the X11 event queue (& no timers pending), then we return
 * immediately.
 */
void
gui_mch_update()
{
    gint pending;

    /* Pitty is that this will block if there are no events
       on the queue. */
    while(((pending = gtk_events_pending()) > 1) && !vim_is_input_buf_full())
	gtk_main_iteration();

    if ((pending == 1) && !vim_is_input_buf_full()) {
	/* Catch up with the last event without blocking on the X11 queue.
	   Yes I know this looks quite crude but at least it works well. */

	gtk_idle_add((GtkFunction) idle_function, main_window);
	gtk_main();
    }

    return;
}

static gint
input_timer_cb(gpointer data)
{
    /* we don't use this information currently */
    int *timed_out = (int *) data;

    *timed_out = TRUE;
    if (gtk_main_level() > 0)
	gtk_main_quit();

    return FALSE;		/* don't happen again */
}

/*
 * GUI input routine called by gui_wait_for_chars().  Waits for a character
 * from the keyboard.
 *  wtime == -1     Wait forever.
 *  wtime == 0      This should never happen.
 *  wtime > 0       Wait wtime milliseconds for a character.
 * Returns OK if a character was found to be available within the given time,
 * or FAIL otherwise.
 */
int
gui_mch_wait_for_chars(long wtime)
{
    int focus;
    guint timer;

    /* this isn't needed in fact */
    static int timed_out = FALSE;

    timed_out = FALSE;

    /* this timeout makes only sure, that, we will at least return
       properly, if the characters didn't arrive */

    if (wtime > 0)
	timer = gtk_timeout_add(wtime, input_timer_cb, &timed_out);
    else
	timer = 0;

    focus = gui.in_focus;

    do {
	/* Stop or start blinking when focus changes */
	if (gui.in_focus != focus) {
	    if (gui.in_focus)
		gui_mch_start_blink();
	    else
		gui_mch_stop_blink();
	    focus = gui.in_focus;
	}
	/* loop in gtk_main() until timeout or input */
	gtk_main();

	/* Got char, return immediately */
	if (!vim_is_input_buf_empty()) {
	    if (timer != 0 && !timed_out) {
		gtk_timeout_remove(timer);
		timer = 0;
	    }
	    return OK;
	}
    } while ((wtime < 0) && !timed_out);


    /*
     * Flush all eventually pending (drawing) events.
     */
    gtk_main_iteration_do(FALSE);

    return FAIL;
}


/****************************************************************************
 * Output drawing routines.
 ****************************************************************************/


/* Flush any output to the screen */
void
gui_mch_flush()
{
    gdk_flush();
}

/*
 * Clear a rectangular region of the screen from text pos (row1, col1) to
 * (row2, col2) inclusive.
 */
void
gui_mch_clear_block(int row1, int col1, int row2, int col2)
{
    GdkGC *gc;
    GdkColor color;

    color.pixel = gui.back_pixel;

    gc = gdk_gc_new(drawing_area->window);
    gdk_gc_set_exposures(gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);
    gdk_gc_set_foreground(gc, &color);

    /*
     * Clear one extra pixel at the right, for when bold characters have
     * spilled over to the next column.  This can erase part of the next
     * character however in the usage context of this function it will be
     * overriden immediately by the correct character again.
     */

    gdk_draw_rectangle(drawing_area->window, gc, TRUE,
		       FILL_X(col1), FILL_Y(row1),
		       (col2 - col1 + 1) * gui.char_width + 1,
		       (row2 - row1 + 1) * gui.char_height);
    gdk_gc_destroy(gc);
}

void
gui_mch_clear_all(void)
{
    gdk_window_clear(drawing_area->window);
}

/*
 * Scroll the text between gui.scroll_region_top & gui.scroll_region_bot by the
 * number of lines given.  Positive scrolls down (text goes up) and negative
 * scrolls up (text goes down).
 */
static void check_copy_area(void)
{
    XEvent event;
    XGraphicsExposeEvent *gevent;

    if (gui.visibility != GDK_VISIBILITY_PARTIAL)
	return;

    gdk_flush();

    /* Wait to check whether the scroll worked or not */
    for (;;) {
	if (XCheckTypedEvent(GDK_DISPLAY(), NoExpose, &event))
	    return;		/* The scroll worked. */

	if (XCheckTypedEvent(GDK_DISPLAY(), GraphicsExpose, &event)) {
	    gevent = (XGraphicsExposeEvent *) & event;
	    gui_redraw(gevent->x, gevent->y, gevent->width, gevent->height);
	    if (gevent->count == 0)
		return;		/* This was the last expose event */
	}
	gdk_flush();
    }
}

/*
 * Delete the given number of lines from the given row, scrolling up any
 * text further down within the scroll region.
 */
void
gui_mch_delete_lines(int row, int num_lines)
{
    if (gui.visibility == GDK_VISIBILITY_FULLY_OBSCURED)
	return;			/* Can't see the window */

    if (num_lines <= 0)
	return;

    if (row + num_lines > gui.scroll_region_bot) {
	/* Scrolled out of region, just blank the lines out */
	gui_clear_block(row, 0, gui.scroll_region_bot, (int) Columns - 1);
    } else {
	GdkGC *gc;

	gc = gdk_gc_new(drawing_area->window);
	gdk_gc_set_exposures(gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);
	gdk_gc_set_foreground(gc, &foreground_color);
	gdk_gc_set_background(gc, &background_color);

	/* copy one extra pixel, for when bold has spilled over */
	gdk_window_copy_area(drawing_area->window, gc,
			     FILL_X(0), FILL_Y(row),
			     drawing_area->window,
			     FILL_X(0), FILL_Y(row + num_lines),
			     gui.char_width * (int) Columns + 1,
			     gui.char_height *
				(gui.scroll_region_bot - row - num_lines + 1));
	gdk_gc_destroy(gc);

	/* Update gui.cursor_row if the cursor scrolled or copied over */
	if (gui.cursor_row >= row) {
	    if (gui.cursor_row < row + num_lines)
		gui.cursor_is_valid = FALSE;
	    else if (gui.cursor_row <= gui.scroll_region_bot)
		gui.cursor_row -= num_lines;
	}
	gui_clear_block(gui.scroll_region_bot - num_lines + 1, 0,
			gui.scroll_region_bot, (int) Columns - 1);
	check_copy_area();
    }
}

/*
 * Insert the given number of lines before the given row, scrolling down any
 * following text within the scroll region.
 */
void
gui_mch_insert_lines(int row, int num_lines)
{
    if (gui.visibility == GDK_VISIBILITY_FULLY_OBSCURED)
	return;			/* Can't see the window */

    if (num_lines <= 0)
	return;

    if (row + num_lines > gui.scroll_region_bot) {
	/* Scrolled out of region, just blank the lines out */
	gui_clear_block(row, 0, gui.scroll_region_bot, (int) Columns - 1);
    } else {
	GdkGC *gc;

	gc = gdk_gc_new(drawing_area->window);
	gdk_gc_set_exposures(gc, gui.visibility != GDK_VISIBILITY_UNOBSCURED);
	gdk_gc_set_foreground(gc, &foreground_color);
	gdk_gc_set_background(gc, &background_color);

	/* copy one extra pixel, for when bold has spilled over */
	gdk_window_copy_area(drawing_area->window, gc,
			     FILL_X(0), FILL_Y(row + num_lines),
			     drawing_area->window,
			     FILL_X(0), FILL_Y(row),
			     gui.char_width * (int) Columns + 1,
			     gui.char_height *
				(gui.scroll_region_bot - row - num_lines + 1));
	gdk_gc_destroy(gc);

	/* Update gui.cursor_row if the cursor scrolled or copied over */
	if (gui.cursor_row >= gui.row) {
	    if (gui.cursor_row <= gui.scroll_region_bot - num_lines)
		gui.cursor_row += num_lines;
	    else if (gui.cursor_row <= gui.scroll_region_bot)
		gui.cursor_is_valid = FALSE;
	}
	gui_clear_block(row, 0, row + num_lines - 1, (int) Columns - 1);
	check_copy_area();
    }
}

/*
 * X Selection stuff, for cutting and pasting text to other windows.
 */
void
clip_mch_request_selection()
{
    gboolean success;

    /* First try to get the content of our own special clipboard. */
    success = gtk_selection_convert(drawing_area,
				    GDK_SELECTION_PRIMARY, clipboard.atom,
				    GDK_CURRENT_TIME);
    gtk_main();		/* wait until signal arrives */

    /* Ok now try to get it out of the usual string selection. */
    success = gtk_selection_convert(drawing_area, GDK_SELECTION_PRIMARY,
				    GDK_TARGET_STRING,
				    GDK_CURRENT_TIME);
    gtk_main();	/* wait until signal arrives */
}

void
clip_mch_lose_selection()
{
    gtk_selection_owner_set(NULL, GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);

    /* I'm not quite sure this is necessary (--mdcki) */
    if (gtk_main_level() > 0)
	gtk_main_quit();
}

/*
 * Check whatever we allready own the selection
 */
int
clip_mch_own_selection()
{
#if 0
    return gdk_selection_owner_get(
				GDK_SELECTION_PRIMARY) == drawing_area->window;
#else
    /* At this point we always already own the clipboard */

    return OK;
#endif
}

/*
 * Send the current selection to the clipboard.
 */
void
clip_mch_set_selection()
{
    gtk_selection_owner_set(drawing_area,
			    GDK_SELECTION_PRIMARY,
			    GDK_CURRENT_TIME);
}


/*
 * Make a menu item appear either active or not active (grey or not grey).
 */
void
gui_mch_menu_grey(GuiMenu * menu, int grey)
{
    if (menu->id == 0)
	return;

    gui_mch_menu_hidden(menu, FALSE);
    gtk_widget_set_sensitive(menu->id, !grey);
}

/*
 * Make menu item hidden or not hidden.
 */
void
gui_mch_menu_hidden(GuiMenu * menu, int hidden)
{
    if (menu->id == 0)
	return;

    if (hidden) {
	if (GTK_WIDGET_VISIBLE(menu->id))
	    gtk_widget_hide(menu->id);
    } else {
	if (!GTK_WIDGET_VISIBLE(menu->id))
	    gtk_widget_show(menu->id);
    }
}

/*
 * This is called after setting all the menus to grey/hidden or not.
 */
void
gui_mch_draw_menubar()
{
    /* just make sure that the visual changes get effect immediately */
    gtk_main_iteration_do(FALSE);
}

/*
 * Scrollbar stuff.
 */
void
gui_mch_enable_scrollbar(GuiScrollbar * sb, int flag)
{
    if (sb->id == 0)
	return;
    if (flag)
	gtk_widget_show(sb->id);
    else
	gtk_widget_hide(sb->id);
}


/*
 * Return the lightness of a pixel.  White is 255.
 *
 * FIXME: port it compleatly to GDK.
 */
int
gui_mch_get_lightness(GuiColor pixel)
{
    XColor xc;
    Colormap colormap;

    colormap = DefaultColormap(GDK_DISPLAY(), XDefaultScreen(GDK_DISPLAY()));
    xc.pixel = pixel;
    XQueryColor(GDK_DISPLAY(), colormap, &xc);

    return (int) (xc.red * 3 + xc.green * 6 + xc.blue) / (10 * 256);
}

/*
 * Return the RGB value of a pixel as "#RRGGBB".
 *
 * FIXME: It should be possible to avoid any direct usage of Xlib. Use gdk
 * instead!  However since the corresponding capabilities seem to first appear
 * in GTK+-1.1.3 we may decide to leave this as it is as well
 */
char_u *
gui_mch_get_rgb(GuiColor pixel)
{
    XColor xc;
    Colormap colormap;
    static char_u retval[10];
    static int prevent_congestion = 0; /* Reduce the race! */


    while(prevent_congestion) ;

    ++prevent_congestion;
    colormap = DefaultColormap(GDK_DISPLAY(), XDefaultScreen(GDK_DISPLAY()));

    xc.pixel = pixel;
    XQueryColor(GDK_DISPLAY(), colormap, &xc);

    sprintf((char *) retval, "#%02x%02x%02x",
	    (unsigned) xc.red >> 8,
	    (unsigned) xc.green >> 8,
	    (unsigned) xc.blue >> 8);
    /* hope the following is atomic enought (99.999%) */
    --prevent_congestion;

    return retval;
}

/*
 * Get current y mouse coordinate in text window.
 * Return -1 when unknown.
 */
int
gui_mch_get_mouse_x(void)
{
    int winx, winy;
    GdkModifierType mask;

    if (gdk_window_get_pointer(drawing_area->window, &winx, &winy, &mask)) {
	if (gui.which_scrollbars[SBAR_LEFT])
	    return winx;
	return winx;
    }
    return -1;
}

int
gui_mch_get_mouse_y(void)
{
    int winx, winy;
    GdkModifierType mask;

    if (gdk_window_get_pointer(drawing_area->window, &winx, &winy, &mask))
	return winy;
    return -1;
}

void
gui_mch_setmouse(int x, int y)
{
    /* Sorry for that, but we can't avoid it, since there seems to be
       no internal GDK mechanism present to accomplish this */
    XWarpPointer(GDK_DISPLAY(), (Window) 0,
		 GDK_WINDOW_XWINDOW(drawing_area->window), 0, 0, 0, 0, x, y);
}


/*** private function defintions ***/

/* redraw the corresponding portions of the screen */
static gint
expose_event(GtkWidget * widget, GdkEventExpose * event)
{
    out_flush();		/* make sure all output has been processed */
    gui_redraw(event->area.x, event->area.y,
	       event->area.width, event->area.height);

    /* Clear the border areas if needed */
    if (event->area.x < FILL_X(0))
	gdk_window_clear_area(drawing_area->window, 0, 0, FILL_X(0), 0);
    if (event->area.y < FILL_Y(0))
	gdk_window_clear_area(drawing_area->window, 0, 0, 0, FILL_Y(0));
    if (event->area.x > FILL_X(Columns))
	gdk_window_clear_area(drawing_area->window,
			      FILL_X((int)Columns), 0, 0, 0);
    if (event->area.y > FILL_Y(Rows))
	gdk_window_clear_area(drawing_area->window, 0, FILL_Y((int)Rows), 0, 0);

    /* gtk_main_quit(); */

    return FALSE;
}


static gint
key_press_event(GtkWidget * widget, GdkEventKey * event, gpointer data)
{
    char_u string[3], string2[3];
    guint key_sym;
    int len;
    int i;
    int modifiers;
    int key;

    key_sym = event->keyval;
    len = event->length;
    g_assert(len <= 3);
    for (i = 0; i < len; ++i)
	string[i] = event->string[i];


    if (key_sym == GDK_space)
	string[0] = ' ';	/* Otherwise Ctrl-Space doesn't work */

    /* Check for Alt/Meta key (Mod1Mask) */
    if (len == 1 && (event->state & GDK_MOD1_MASK)
	&& !(key_sym == GDK_BackSpace || key_sym == GDK_Delete)) {
	/* Don't do this for <S-M-Tab>, that should become K_S_TAB with ALT. */
	if (!(key_sym == GDK_Tab && (event->state & GDK_SHIFT_MASK)))
	    string[0] |= 0x80;
    }

    /* Check for special keys, making sure BS and DEL are recognised. */
    if (len == 0 || key_sym == GDK_BackSpace || key_sym == GDK_Delete) {
	for (i = 0; special_keys[i].key_sym != 0; i++) {
	    if (special_keys[i].key_sym == key_sym) {
		string[0] = CSI;
		string[1] = special_keys[i].code0;
		string[2] = special_keys[i].code1;
		len = 3;
		break;
	    }
	}
    }

    if (len == 0)   /* Unrecognised key */
	return TRUE;

    /* Special keys (and a few others) may have modifiers */
    if (len == 3 || key_sym == GDK_space || key_sym == GDK_Tab ||
	key_sym == GDK_Return || key_sym == GDK_Linefeed ||
	key_sym == GDK_Escape) {
	modifiers = 0;
	if (event->state & GDK_SHIFT_MASK)
	    modifiers |= MOD_MASK_SHIFT;
	if (event->state & GDK_CONTROL_MASK)
	    modifiers |= MOD_MASK_CTRL;
	if (event->state & GDK_MOD1_MASK)
	    modifiers |= MOD_MASK_ALT;

	/*
	 * For some keys a shift modifier is translated into another key code.
	 * Do we need to handle the case where len != 1 and string[0] != CSI?
	 */
	if (string[0] == CSI && len == 3)
	    key = TO_SPECIAL(string[1], string[2]);
	else
	    key = string[0];

	key = simplify_key(key, &modifiers);
	if (IS_SPECIAL(key)) {
	    string[0] = CSI;
	    string[1] = K_SECOND(key);
	    string[2] = K_THIRD(key);
	    len = 3;
	} else {
	    string[0] = key;
	    len = 1;
	}

	if (modifiers) {
	    string2[0] = CSI;
	    string2[1] = KS_MODIFIER;
	    string2[2] = modifiers;
	    add_to_input_buf(string2, 3);
	}
    } /* special keys */

    if (len == 1 && (string[0] == Ctrl('C') || string[0] == intr_char)) {
	trash_input_buf();
	got_int = TRUE;
    }
    add_to_input_buf(string, len);

    /* blank out the pointer if necessary */
    if (p_mh)
	gui_mch_mousehide(TRUE);

    if (gtk_main_level() > 0)
	gtk_main_quit();

    return TRUE;
}

static guint mouse_click_timer = 0;
static int mouse_timed_out = TRUE;

/*
 * Timer used to recognize multiple clicks of the mouse button
 */
static gint
mouse_click_timer_cb(gpointer data)
{
    /* we don't use this information currently */
    int *timed_out = (int *) data;

    *timed_out = TRUE;
    return FALSE;		/* don't happen again */
}

/*
 * Mouse button handling. Note please that we are capturing multiple click's by
 * our own timeout mechanis instead of the one provided by GTK+ istelf. This is
 * due to the way the generic VIM code is recognizing multiple clicks.
 */
static gint
button_press_event(GtkWidget * widget, GdkEventButton * event)
{
    int button;
    int repeated_click = FALSE;
    int x, y;
    int_u vim_modifiers;

    /*
     * Don't let additional events about multiple clicks send by GTK to us
     * after the inital button press event confuse us.
     */
    if (event->type != GDK_BUTTON_PRESS)
	return FALSE;

    x = event->x;
    y = event->y;


    /* Handle multiple clicks */
    if (!mouse_timed_out && mouse_click_timer) {
	gtk_timeout_remove(mouse_click_timer);
	mouse_click_timer = 0;
	repeated_click = TRUE;
    }

    mouse_timed_out = FALSE;
    mouse_click_timer = gtk_timeout_add(p_mouset,
				  mouse_click_timer_cb, &mouse_timed_out);

    switch (event->button) {
    case 1:
	button = MOUSE_LEFT;
	break;
    case 2:
	button = MOUSE_MIDDLE;
	break;
    case 3:
	button = MOUSE_RIGHT;
	break;
    default:
	return FALSE;		/* Unknown button */
    }

    vim_modifiers = 0x0;
    if (event->state & GDK_SHIFT_MASK)
	vim_modifiers |= MOD_MASK_SHIFT;
    if (event->state & GDK_CONTROL_MASK)
	vim_modifiers |= MOD_MASK_CTRL;
    if (event->state & GDK_MOD1_MASK)
	vim_modifiers |= MOD_MASK_ALT;

    gui_send_mouse_event(button, x, y, repeated_click, vim_modifiers);
    if (gtk_main_level() > 0)
	gtk_main_quit();	/* make sure the above will be handled immediately */

    return TRUE;
}

static guint motion_repeat_timer = 0;

/*
 * Timer used to recognize multiple clicks of the mouse button
 */
static gint
motion_repeat_timer_cb(gpointer data)
{
    gint x, y;
    GdkModifierType state;
    GdkEventMotion event;

    gdk_window_get_pointer(drawing_area->window, &x, &y, &state);

    if (!(state & (GDK_BUTTON1_MASK | GDK_BUTTON3_MASK)))
	return FALSE;

    /* Fake a motion event. */
    event.is_hint = 0;
    event.x = x;
    event.y = y;
    event.state = state;

    /* FIXME: argh! for some unknown reason this doesn't lead to
      nice autoscrolling and automatic extension of the selection area.
      However the autorepeating works nice already.
      We need certainly to investigate this further.

      BTW: It doesn't work properly in the motif version too. So maybe we
      need to look at how it's done in the Windows version again...

      I'm assuming that this is somehow related to the gui_mouse_moved stuff...
    */

    motion_notify_event(drawing_area, &event);

    /* Don't happen again. We will get reinstalled in the synthetic
       event if needed - thus repeating will still work. */

    return FALSE;
}


static gint
button_release_event(GtkWidget * widget, GdkEventButton * event)
{
    int x, y;
    int_u vim_modifiers;

    /* Remove any motion "mashine gun" timers used for automatic further
       extension of allocation areas if outside of the applications window
       area .*/
    if (motion_repeat_timer) {
	gtk_timeout_remove(motion_repeat_timer);
	motion_repeat_timer = 0;
    }

    x = event->x;
    y = event->y;

    vim_modifiers = 0x0;
    if (event->state & GDK_SHIFT_MASK)
	vim_modifiers |= MOD_MASK_SHIFT;
    if (event->state & GDK_CONTROL_MASK)
	vim_modifiers |= MOD_MASK_CTRL;
    if (event->state & GDK_MOD1_MASK)
	vim_modifiers |= MOD_MASK_ALT;

    gui_send_mouse_event(MOUSE_RELEASE, x, y, FALSE, vim_modifiers);
    if (gtk_main_level() > 0)
	gtk_main_quit();		/* make sure it will be handled immediately */

    return TRUE;
}


static gint
motion_notify_event(GtkWidget * widget, GdkEventMotion * event)
{
    gint x, y;
    GdkModifierType state;
    int_u vim_modifiers;
    int button;

    if (event->is_hint)
	gdk_window_get_pointer(event->window, &x, &y, &state);
    else {
	x = event->x;
	y = event->y;
	state = event->state;
    }
    button = (event->state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK |
			      GDK_BUTTON3_MASK | GDK_BUTTON4_MASK |
			      GDK_BUTTON5_MASK))
	      ? MOUSE_DRAG : ' ';

    /* If our pointer is currently hidden, then we should show it. */
    gui_mch_mousehide(FALSE);

    /* Just moving the rodent above the drawing area without
       any button beeing pressed. */
    if (button != MOUSE_DRAG) {
	gui_mouse_moved(y);
	return TRUE;
    }

    vim_modifiers = 0x0;
    if (state & GDK_SHIFT_MASK)
	vim_modifiers |= MOD_MASK_SHIFT;
    if (state & GDK_CONTROL_MASK)
	vim_modifiers |= MOD_MASK_CTRL;
    if (state & GDK_MOD1_MASK)
	vim_modifiers |= MOD_MASK_ALT;

    /* inform the editor egine about the occurence of this event */
    gui_send_mouse_event(button, x, y, FALSE, vim_modifiers);
    if (gtk_main_level() > 0)
	gtk_main_quit();

    /*
     * Auto repeat timer handling.
     */
    if (x < 0 || y < 0
       || x >= drawing_area->allocation.width
       || y >= drawing_area->allocation.height) {

	/* just in case remove stale timers */
	if (motion_repeat_timer)
	    gtk_timeout_remove(motion_repeat_timer);
	/* shoot again */
	motion_repeat_timer = gtk_timeout_add(100,
					  motion_repeat_timer_cb, NULL);
    }

    return TRUE; /* handled */
}

/*
 * Function called when window alread closed.
 * We can't do much more here than to trying to preserve what had been done,
 * since the window is already inevitably going away.
 */
static void
destroy_callback(void)
{
    /* preserve files and exit */
    preserve_exit();
    if (gtk_main_level() > 0)
	gtk_main_quit();
}

#ifdef GTK_HAVE_FEATURES_1_1_0
/*
 * Get a font structure for highlighting.
 */
static void
font_selection_ok(GtkWidget * w, GtkFontSelectionDialog * fs)
{
    if (selected_font_name)
	g_free(selected_font_name);

    selected_font_name = (char_u *)g_strdup(
				gtk_font_selection_dialog_get_font_name(fs));
    gtk_widget_destroy(GTK_WIDGET(fs));
}

static void
font_selection_destroy(GtkWidget * w, gpointer * data)
{
    if (gtk_main_level() > 0)
	gtk_main_quit();
}
#endif

